import math
import ptn_data
from typing import List
from typing import Dict


class LinePool:
    """
    Class representing the line pool.
    """

    def __init__(self) -> None:
        """
        Initialization of an empty line pool.
        """
        self.pool = []
        self.max_id = 0;

    def get_lines(self) -> List["Line"]:
        return self.pool

    def add_line(self, line: "Line") -> None:
        self.pool.append(line)
        if line.get_directed_line_id() > self.max_id:
            self.max_id = line.get_directed_line_id()


class Line:
    """
    Class representing a line.
    """

    def __init__(self, directed_line_id: int, pool: LinePool, undirected_line_id: int = -1, frequency: int = -1,
                 repetition: int = 1) -> None:
        """
        Initialization of a line.
        :param directed_line_id: directed line id. Can be negative, if the line was only given in one direction
                and the line represents the backwards direction.
        :param pool: corresponding line pool
        :param undirected_line_id: undirected line id. Forward and backward direction have the same undirected line id
        :param frequency: frequency of the line
        :param repetition: frequency repetition. If a line frequency is 2, two lines with repetion 1 and 2,
                respectively, are created.
        """
        self.length = 0
        self.cost = 0
        self.edges = []
        self.directed_line_id = directed_line_id
        if undirected_line_id != -1:
            self.undirected_line_id = undirected_line_id
        else:
            self.undirected_line_id = directed_line_id
        self.first_stop = None
        self.last_stop = None
        self.frequency = frequency
        self.repetition = repetition
        pool.add_line(self)

    def add_edge(self, edge: ptn_data.PtnEdge, directed: bool = True, forward: bool = True) -> None:
        if directed:
            if not self.edges:
                self.edges.append(edge)
            elif self.edges[- 1].get_right_node().get_node_id() != edge.get_left_node().get_node_id():
                raise RuntimeError("Edge %d does not fit in line %d!" % (edge.get_edge_id(), self.directed_line_id))
            else:
                self.edges.append(edge)
                self.last_stop = edge.get_right_node()
        else:
            if not self.edges:
                if forward:
                    self.edges.append(edge)
                else:
                    self.edges.append(edge.get_backward_edge())
            else:
                if self.edges.__len__() == 1:
                    if forward and self.edges[0].get_right_node().get_node_id() not in \
                            [edge.get_left_node().get_node_id(), edge.get_right_node().get_node_id()]:
                        first_edge = self.edges.pop()
                        self.edges.append(first_edge.get_backward_edge())
                    elif not forward and self.edges[0].get_left_node().get_node_id() not in \
                            [edge.get_left_node().get_node_id(), edge.get_right_node().get_node_id()]:
                        first_edge = self.edges.pop()
                        self.edges.append(first_edge.get_backward_edge())
                if forward and self.edges[- 1].get_right_node().get_node_id() == edge.get_left_node().get_node_id():
                    self.edges.append(edge)
                elif forward and self.edges[- 1].get_right_node().get_node_id() == edge.get_right_node().get_node_id():
                    self.edges.append(edge.get_backward_edge())
                elif not forward and self.edges[0].get_left_node().get_node_id() == edge.get_right_node().get_node_id():
                    self.edges.insert(0, edge)
                elif not forward and self.edges[0].get_left_node().get_node_id() == edge.get_left_node().get_node_id():
                    self.edges.insert(0, edge.get_backward_edge())
                else:
                    raise RuntimeError("Edge %d does not fit in line %d!" % (edge.get_edge_id(), self.directed_line_id))

    def get_undirected_line_id(self) -> int:
        return self.undirected_line_id

    def get_directed_line_id(self) -> int:
        return self.directed_line_id

    def get_edges(self) -> List[ptn_data.PtnEdge]:
        return self.edges

    def get_nodes(self) -> List[ptn_data.PtnNode]:
        nodes = []
        nodes.append(self.get_first_stop())
        for edge in self.edges:
            nodes.append(edge.get_right_node())
        return nodes

    def to_string(self) -> str:
        if self.directed_line_id > 0:
            direction = ">"
        else:
            direction = "<"
        return "(%d,%s,%d/%d)" % (self.undirected_line_id, direction, self.repetition, self.frequency)

    def get_first_stop(self) -> ptn_data.PtnNode:
        if not self.edges:
            return None
        return self.edges[0].get_left_node()

    def get_last_stop(self) -> ptn_data.PtnNode:
        if not self.edges:
            return None
        return self.edges[-1].get_right_node()

    def get_frequency(self) -> int:
        return self.frequency

    def get_repetition(self) -> int:
        return self.repetition

    def get_direction(self) -> str:
        if self.directed_line_id > 0:
            return ">"
        else:
            return "<"

    def __eq__(self, other: object) -> bool:
        if other is None:
            return False
        if not isinstance(other, Line):
            return False
        return self.directed_line_id == other.get_directed_line_id() \
                   and self.repetition == other.get_repetition()

    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)

    def __hash__(self) -> int:
        return hash((self.directed_line_id, self.repetition))

