import ptn_data
from typing import List


class OD:
    """
    Class representing the OD matrix.
    """

    def __init__(self) -> None:
        """
        Initialisation of an empty OD matrix.
        """
        self.od_pairs = []

    def add_od_pair(self, od_pair: "ODPair") -> None:
        self.od_pairs.append(od_pair)

    def get_active_od_pairs(self) -> List["ODPair"]:
        od_pairs = []
        for od_pair in self.od_pairs:
            if od_pair.is_active():
                od_pairs.append(od_pair)
        return od_pairs

    def get_all_od_pairs(self) -> List["ODPair"]:
        return self.od_pairs

    def get_od_pair(self, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode) -> "ODPair":
        for od_pair in self.od_pairs:
            if od_pair.get_origin() == origin and od_pair.get_destination() == destination:
                return od_pair
        raise RuntimeError('Requested OD-pair (%d,%d) not in list.' % (origin.get_node_id(), destination.get_node_id()))

    def get_n_time_slices(self, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode) -> int:
        for od_pair in self.od_pairs:
            if od_pair.get_origin() == origin and od_pair.get_destination() == destination:
                return od_pair.get_n_time_slices()
        raise RuntimeError('Requested OD-pair (%d,%d) not in list.' % (origin.get_node_id(), destination.get_node_id()))

    def get_max_travel_time(self, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode) -> int:
        for od_pair in self.od_pairs:
            if od_pair.get_origin() == origin and od_pair.get_destination() == destination:
                return od_pair.get_max_travel_time()
        raise RuntimeError('Requested OD-pair (%d,%d) not in list.' % (origin.get_node_id(), destination.get_node_id()))

    def get_min_turn_around_time(self, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode) -> int:
        for od_pair in self.od_pairs:
            if od_pair.get_origin() == origin and od_pair.get_destination() == destination:
                return od_pair.get_min_turn_around_time()
        raise RuntimeError('Requested OD-pair (%d,%d) not in list.' % (origin.get_node_id(), destination.get_node_id()))

    def get_min_turn_around_distance(self, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode) -> float:
        for od_pair in self.od_pairs:
            if od_pair.get_origin() == origin and od_pair.get_destination() == destination:
                return od_pair.get_min_turn_around_distance()
        raise RuntimeError('Requested OD-pair (%d,%d) not in list.' % (origin.get_node_id(), destination.get_node_id()))

    def get_n_passengers(self, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode, time_slice: int) -> int:
        for od_pair in self.od_pairs:
            if od_pair.get_origin() == origin and od_pair.get_destination() == destination:
                return od_pair.get_n_passengers(time_slice)
        raise RuntimeError('Requested OD-pair (%d,%d) not in list.' % (origin.get_node_id(), destination.get_node_id()))


class ODPair:
    """
    Class representing an OD pair.
    """

    def __init__(self, od: OD, origin: ptn_data.PtnNode, destination: ptn_data.PtnNode, min_turn_around_time: int,
                 min_turn_around_distance: float = 0, n_time_slices: int = -1, max_travel_time: int = -1,
                 active: bool = True) -> None:
        """
        Initialization of an Od pair.
        :param od: corresponding OD matrix
        :param origin: PTN node which is the origin of the OD pair
        :param destination: PTN node which is the destination of the OD pair
        :param min_turn_around_time: minimal tun around time between the OD nodes
        :param min_turn_around_distance: minimal turn around distance between the OD nodes
        :param n_time_slices: number of time slices considered for the OD pair
        :param max_travel_time: maximum travel time for the OD pair
        :param active: whether the OD pair is active, i.e., iif it is routed during the optimization process
        """
        self.origin = origin
        self.destination = destination
        self.n_time_slices = n_time_slices
        self.max_travel_time = max_travel_time
        self.min_turn_around_time = min_turn_around_time
        self.min_turn_around_distance = min_turn_around_distance
        self.active = active
        self.passengers = {}
        od.add_od_pair(self)

    def get_n_time_slices(self) -> int:
        return self.n_time_slices

    def get_origin(self) -> ptn_data.PtnNode:
        return self.origin

    def get_destination(self) -> ptn_data.PtnNode:
        return self.destination

    def get_max_travel_time(self) -> int:
        return self.max_travel_time

    def get_min_turn_around_time(self) -> int:
        return self.min_turn_around_time

    def get_min_turn_around_distance(self) -> float:
        return self.min_turn_around_distance

    def get_n_passengers(self, time_slice: int) -> int:
        return self.passengers[time_slice]

    def is_active(self) -> bool:
        n_passengers = 0
        for time_slice in self.passengers.keys():
            n_passengers += self.passengers[time_slice]
        return self.active and n_passengers > 0

    def set_n_passengers(self, time_slice: int, n_passengers: int) -> None:
        self.passengers[time_slice] = n_passengers

    def get_total_passengers(self) -> int:
        total_passengers = 0
        for t in range(1, self.n_time_slices + 1):
            total_passengers += self.passengers[t]
        return total_passengers

    def get_penalty(self, time_1: int, time_2: int) -> float:
        return abs(time_1 - time_2) / self.n_time_slices

    def to_string(self) -> str:
        return "(" + self.origin.to_string() + "," + self.destination.to_string() + ")"
