from typing import ValuesView
from typing import List


class Ptn:
    """
    Class representing a Public Transportation Network (PTN).
    """

    def __init__(self) -> None:
        """
        Initialisation of an empty PTN.
        """
        self.nodes = {}

    def add_node(self, node: "PtnNode") -> None:
        self.nodes[node.get_node_id()] = node

    def get_nodes(self) -> ValuesView["PtnNode"]:
        return self.nodes.values()

    def get_node(self, node_id: int) -> "PtnNode":
        return self.nodes[node_id]

    def get_edges(self) -> List["PtnEdge"]:
        edges = []
        for node in self.nodes.values():
            for edge in node.get_incident_edges():
                if not edges.__contains__(edge):
                    edges.append(edge)
        return edges

    def get_edge(self, edge_id: int) -> "PtnEdge":
        for node in self.nodes.values():
            for edge in node.get_incident_edges():
                if edge.get_edge_id() == edge_id:
                    return edge
        raise RuntimeError('Edge with index %d not found.' % edge_id)


class PtnNode:
    """
    Class representing a node in a PTN, i.e., a station or stop.
    """

    def __init__(self, ptn: Ptn, node_id: int, min_wait_time: int, max_wait_time: int, min_trans_time: int,
                 max_trans_time: int, x_coordinate: float, y_coordinate: float, min_from_depot_time: int = 0,
                 min_to_depot_time: int = 0, min_from_depot_distance: float = 0, min_to_depot_distance: float = 0) -> None:
        """
        Initialization of a PTN node.
        :param ptn: corresponding PTN
        :param node_id: id of th enode
        :param min_wait_time: minimal waiting time in the node
        :param max_wait_time: maximal waiting time in the node
        :param min_trans_time: minimal transfer time at the node
        :param max_trans_time: maximal transfer time at the node
        :param x_coordinate: x-coordinate
        :param y_coordinate: y-coordinate
        :param min_from_depot_time: minimal time it takes to get from the depot to the node
        :param min_to_depot_time: maximal time it takes to get from the node to the node
        :param min_from_depot_distance: minimal distance of a path from the depot to the node
        :param min_to_depot_distance: minimal distance of a path from the node to the depot
        """
        self.node_id = node_id
        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.min_wait_time = min_wait_time
        self.max_wait_time = max_wait_time
        self.min_trans_time = min_trans_time
        self.max_trans_time = max_trans_time
        self.min_from_depot_time = min_from_depot_time
        self.min_to_depot_time = min_to_depot_time
        self.min_from_depot_distance = min_from_depot_distance
        self.min_to_depot_distance = min_to_depot_distance
        self.incident_edges = []
        ptn.add_node(self)

    def add_edge(self, edge: "PtnEdge") -> None:
        if not self.incident_edges.__contains__(edge):
            self.incident_edges.append(edge)

    def get_incident_edges(self) -> List["PtnEdge"]:
        return self.incident_edges

    def get_node_id(self) -> int:
        return self.node_id

    def get_min_wait_time(self) -> int:
        return self.min_wait_time

    def get_max_wait_time(self) -> int:
        return self.max_wait_time

    def get_min_trans_time(self) -> int:
        return self.min_trans_time

    def get_max_trans_time(self) -> int:
        return self.max_trans_time

    def to_string(self) -> str:
        return str(self.node_id)

    def equals(self, other_node: "PtnNode") -> bool:
        if not isinstance(other_node, PtnNode):
            return False
        return self.node_id == other_node.get_node_id()


class PtnEdge:
    """
    Class representing an edge in the PTN, i.e., a direction connection between to nodes, e.g. a track
    """

    def __init__(self, edge_id: int, left_node: PtnNode, right_node: PtnNode, min_drive_time: int, max_drive_time: int,
                 directed: bool = True, backward_edge: "PtnEdge" = None) -> None:
        """
        Initialization of a PTN edge.
        :param edge_id: id of the edge
        :param left_node: source node of the edge
        :param right_node: target node of the edge
        :param min_drive_time: minimal time it takes to drive the edge
        :param max_drive_time: maximum time that is allowed for driving on the edge
        :param directed: whether the edge is directed
        :param backward_edge: backward edge if the edge is directed
        """
        self.edge_id = edge_id
        self.left_node = left_node
        self.right_node = right_node
        self.min_drive_time = min_drive_time
        self.max_drive_time = max_drive_time
        self.backward_edge = backward_edge
        self.min_frequency = 0
        self.max_frequency = 0
        left_node.add_edge(self)
        if not directed and backward_edge is None:
            self.backward_edge = PtnEdge(-edge_id, right_node, left_node, min_drive_time, max_drive_time, True, self)

    def get_edge_id(self) -> int:
        return self.edge_id

    def get_left_node(self) -> PtnNode:
        return self.left_node

    def get_right_node(self) -> PtnNode:
        return self.right_node

    def get_min_drive_time(self) -> int:
        return self.min_drive_time

    def get_max_drive_time(self) -> int:
        return self.max_drive_time

    def get_backward_edge(self) -> "PtnEdge":
        return self.backward_edge

    def to_string(self) -> str:
        return str(
            abs(self.edge_id)) + "=(" + self.left_node.to_string() + "," + self.get_right_node().to_string() + ")"
